#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) Metaswitch Networks.
"""ChangeLogRenderer for mistletoe."""

import collections
import html
import logging
from mistletoe.base_renderer import BaseRenderer

log = logging.getLogger(__name__)


ListEntry = collections.namedtuple("ListEntry", ["depth", "number", "content"])


class ListCounter(object):
    def __init__(self, start):
        self.current = start

    def __next__(self):
        current = self.current
        if current is not None:
            self.current += 1
        return current


class ChangeLogRenderer(BaseRenderer):
    def __init__(self, version: str, *extras):
        super().__init__(*extras)
        self.version = version
        self.diff_url = None

    def __exit__(self, *args):
        super().__exit__(*args)
        self.diff_url = None

    def render_to_plaintext(self, token):
        if hasattr(token, "children"):
            rendered = [self.render_to_plaintext(child) for child in token.children]
            return "".join(rendered)
        else:
            return token.content

    def render_document(self, token):
        to_render = []
        rendering = False

        for child in token.children:
            if child.__class__.__name__ == "Heading" and child.level == 2:
                # Get the text of the first child of this heading. This should be the
                # version number, or "Unreleased".
                heading_text = self.render_to_plaintext(child.children[0])

                # Only render things under the right level 2 heading.
                if heading_text == self.version:
                    rendering = True
                    if child.children[0].__class__.__name__ == "Link":
                        self.diff_url = child.children[0].target

                else:
                    rendering = False

            if rendering:
                to_render.append(child)
            else:
                log.debug("Not rendering %s", child)

        if to_render and to_render[-1].__class__.__name__ == "Heading":
            # The last field is a heading. Headings on their own are usually because
            # people haven't deleted the Changed or Added heading. Rather than render
            # this, let's just delete it.
            log.warning(
                "Deleting empty heading as is the last field: %s",
                self.render_to_plaintext(to_render[-1]),
            )
            to_render.pop()

        log.debug("Document contents %r", to_render)

        rendered = [self.render(child) for child in to_render]
        return "".join(rendered)

    def render(self, token):
        ret = self.render_map[token.__class__.__name__](token)
        log.debug("Rendering %r returns %r", token, ret)
        return ret

    def render_strong(self, token):
        return "*{}*".format(self.render_inner(token))

    def render_emphasis(self, token):
        return "_{}_".format(self.render_inner(token))

    def render_inline_code(self, token):
        inner = token.children[0].content
        return "`{}`".format(inner)

    def render_strikethrough(self, token):
        return "~{}~".format(self.render_inner(token))

    def render_image(self, token):
        return "<{}|{}>".format(token.src, self.escape_html(token.src))

    def render_link(self, token):
        template = "<{target}|{inner}>"
        target = token.target
        inner = self.escape_html(self.render_to_plaintext(token))
        return template.format(target=target, inner=inner)

    def render_auto_link(self, token):
        template = "<{target}|{inner}>"
        if token.mailto:
            target = "mailto:{}".format(token.target)
        else:
            target = token.target
        inner = self.escape_html(self.render_to_plaintext(token))
        return template.format(target=target, inner=inner)

    def render_escape_sequence(self, token):
        return self.render_inner(token)

    def render_raw_text(self, token):
        return self.escape_html(token.content)

    @staticmethod
    def render_html_span(token):
        return token.content

    def render_heading(self, token):
        template = "*{}*\n" if token.level == 3 else "{}\n"
        return template.format(self.render_to_plaintext(token))

    def render_quote(self, token):
        inner = self.render_to_inner(token)
        return "> {}\n".format(inner)

    def render_paragraph(self, token):
        return self.render_inner(token)

    def render_block_code(self, token):
        template = "```\n{inner}```\n"
        inner = token.children[0].content
        return template.format(inner=inner)

    def render_list(self, token):
        analysed = self.analyse_list(token, 0)
        rendered_list = [self.render_listentry(e) for e in analysed]
        return "".join(rendered_list)

    def render_listentry(self, listentry: ListEntry) -> str:
        if listentry.number is not None:
            bullet = "{}.".format(listentry.number)
        else:
            bullet = "\u2022"

        leading_spaces = " " * (listentry.depth * 2)

        return "{spaces}{bullet} {content}\n".format(
            spaces=leading_spaces, bullet=bullet, content=listentry.content
        )

    def analyse_list(self, token, depth):
        analysed = []

        counter = ListCounter(token.start)
        for list_item in token.children:
            analysed.extend(self.analyse_listitem(list_item, depth, counter))

        return analysed

    def analyse_listitem(self, token, depth, counter):
        entries = []

        for listitem_child in token.children:
            if listitem_child.__class__.__name__ == "List":
                sub_entries = self.analyse_list(listitem_child, depth + 1)
                entries.extend(sub_entries)
            else:
                number = next(counter)
                entries.append(ListEntry(depth, number, self.render(listitem_child)))

        log.debug("Listitem entries: %s", entries)
        return entries

    def render_list_item(self, token):
        return "list_item_uncalled"

    def render_table(self, token):
        return "table_unsupported"

    def render_table_row(self, token, is_header=False):
        return "table_row_unsupported"

    def render_table_cell(self, token, in_header=False):
        return "table_cell_unsupported"

    @staticmethod
    def render_thematic_break(token):
        return "---\n"

    @staticmethod
    def render_line_break(token):
        return "\n"

    @staticmethod
    def render_html_block(token):
        return token.content

    @staticmethod
    def escape_html(raw):
        return html.escape(raw, quote=False)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['announcer']

package_data = \
{'': ['*']}

install_requires = \
['mistletoe>=0.7.1,<0.8.0', 'requests>=2.19,<3.0', 'typing>=3.6.4,<4.0.0']

entry_points = \
{'console_scripts': ['announce = announcer.__init__:main']}

setup_kwargs = {
    'name': 'announcer',
    'version': '2.1.0',
    'description': 'Announce changes in keepachangelog-style CHANGELOG.md files to Slack',
    'long_description': '# announcer\n\nThis tool:\n* takes an [keepachangelog](https://keepachangelog.com/en/1.0.0/)-style CHANGELOG.md file\n* extracts all changes for a particular version\n* and sends a formatted message to a Slack webhook. \n\n## Installation\n\nInstall this tool using pip:\n\n```\npip install announcer\n```\n\n## Tool usage\n\n```\nusage: announce [-h] --slackhook SLACKHOOK --changelogversion CHANGELOGVERSION\n                --changelogfile CHANGELOGFILE --projectname PROJECTNAME\n                [--username USERNAME]\n                [--iconurl ICONURL | --iconemoji ICONEMOJI]\n\nAnnounce CHANGELOG changes on Slack\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --slackhook SLACKHOOK\n                        The incoming webhook URL\n  --changelogversion CHANGELOGVERSION\n                        The changelog version to announce (e.g. 1.0.0)\n  --changelogfile CHANGELOGFILE\n                        The file containing changelog details (e.g.\n                        CHANGELOG.md)\n  --projectname PROJECTNAME\n                        The name of the project to announce (e.g. announcer)\n  --username USERNAME   The username that the announcement will be made as\n                        (e.g. qs-announcer)\n  --iconurl ICONURL     A URL to use for the user icon in the announcement\n  --iconemoji ICONEMOJI\n                        A Slack emoji code to use for the user icon in the\n                        announcement (e.g. party_parrot)\n```\n',
    'author': 'Max Dymond',
    'author_email': 'max.dymond@metaswitch.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) Metaswitch Networks.
"""A tool for announcing keepachangelog format logs to Slack channels"""

import argparse
import json
import logging
import re
import sys
import typing

from .changelogrenderer import ChangeLogRenderer
import mistletoe
import requests

log = logging.getLogger(__name__)


DIFF_URL_RE = re.compile("^(.*)/compare/[^/]+$")


def announce(
    slackhook: str,
    changelogversion: str,
    changelogfile: str,
    projectname: str,
    username: str = None,
    icon_url: str = None,
    icon_emoji: str = None,
):
    """Main script for announce

    """
    # Get the changelog
    log.info("Querying changelog %s", changelogfile)
    changelog = Changelog(changelogfile)

    # Get the version information
    log.info("Getting version %s info from changelog", changelogversion)
    (changelog_info, diff_url) = changelog.get_version_details(changelogversion)

    # Announce the information to Slack.
    log.info("Announcing information to Slack")
    log.debug("Slack hook %s", slackhook)

    # Try and derive the base URL from the diff URL if it exists.
    base_url = None
    if diff_url:
        m = DIFF_URL_RE.match(diff_url)
        if m:
            base_url = m.group(1)

    # Make a message attachment.
    pretext = ["*{0} {1}*".format(projectname, changelogversion)]

    if base_url:
        pretext.append("({0})".format(base_url))

    attachments = [
        {"color": "good", "pretext": " ".join(pretext), "text": changelog_info}
    ]  # type: typing.List[typing.Dict[str, typing.Any]]

    fallback = []
    actions = []

    if diff_url:
        # Add a button to view the changes at the diff URL
        fallback.append("View changes at {0}".format(diff_url))
        actions.append({"type": "button", "text": "View Changes", "url": diff_url})

    if base_url:
        # Add a button to view the CHANGELOG.md.
        changelog_url = "{0}/blob/{1}/CHANGELOG.md".format(base_url, changelogversion)
        fallback.append("View CHANGELOG.md at {0}".format(changelog_url))
        actions.append(
            {"type": "button", "text": "View CHANGELOG.md", "url": changelog_url}
        )

    if actions:
        attachments.append(
            {"fallback": "\n".join(fallback), "color": "good", "actions": actions}
        )

    # Construct the data to send to the endpoint.
    slack_data = {"attachments": attachments}  # type: typing.Dict[str, typing.Any]

    if username:
        slack_data["username"] = username

    if icon_url:
        slack_data["icon_url"] = icon_url
    elif icon_emoji:
        # Wrap the emoji name in colons to use that emoji
        slack_data["icon_emoji"] = ":{}:".format(icon_emoji)

    log.debug("Sending info %s", slack_data)

    # Send the data using requests to the slack hook.
    r = requests.post(
        slackhook, json.dumps(slack_data), headers={"Content-Type": "application/json"}
    )
    r.raise_for_status()

    return ScriptRC.SUCCESS


class Changelog(object):
    def __init__(self, filename: str) -> None:
        self.filename = filename

    def get_version_details(
        self, version: str
    ) -> typing.Tuple[str, typing.Optional[str]]:
        with open(self.filename, "r") as f:
            document = mistletoe.Document(f)

            with ChangeLogRenderer(version) as renderer:
                rendered = renderer.render(document)
                diff_url = renderer.diff_url

        log.debug("Diff URL: %s", diff_url)
        return rendered, diff_url


def main():
    """Main handling function."""

    # Set up basic logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)-5.5s %(message)s",
        stream=sys.stdout,
        level=logging.INFO,
    )

    # Run main script.
    parser = argparse.ArgumentParser(description="Announce CHANGELOG changes on Slack")
    parser.add_argument(
        "--slackhook", dest="slackhook", required=True, help="The incoming webhook URL"
    )
    parser.add_argument(
        "--changelogversion",
        dest="changelogversion",
        required=True,
        help="The changelog version to announce (e.g. 1.0.0)",
    )
    parser.add_argument(
        "--changelogfile",
        dest="changelogfile",
        required=True,
        help="The file containing changelog details (e.g. CHANGELOG.md)",
    )
    parser.add_argument(
        "--projectname",
        dest="projectname",
        required=True,
        help="The name of the project to announce (e.g. announcer)",
    )
    parser.add_argument(
        "--username",
        dest="username",
        help="The username that the announcement will be made as (e.g. qs-announcer)",
    )

    icons = parser.add_mutually_exclusive_group()
    icons.add_argument(
        "--iconurl",
        dest="iconurl",
        help="A URL to use for the user icon in the announcement",
    )
    icons.add_argument(
        "--iconemoji",
        dest="iconemoji",
        help="A Slack emoji code to use for the user icon in the announcement "
        "(e.g. party_parrot)",
    )

    args = parser.parse_args()

    try:
        exit_code = announce(
            args.slackhook,
            args.changelogversion,
            args.changelogfile,
            args.projectname,
            username=args.username,
            icon_url=args.iconurl,
            icon_emoji=args.iconemoji,
        )
    except Exception as e:
        log.exception(e)
        exit_code = ScriptRC.EXCEPTION

    log.info("Returning %d", exit_code)
    return exit_code


class ScriptRC(object):
    SUCCESS = 0
    FAILURE = 1
    EXCEPTION = 2


if __name__ == "__main__":
    sys.exit(main())
